/*
 * Process Variable Library with JVisu Socket Server
 *
 * Copyright (C) 2003 Thomas Gleixner <tglx@linutronix.de>
 *           (C) 2005 Robert Schwebel <r.schwebel@pengutronix.de>
 *
 * Copyright 2005 Pengutronix
 *
 * Please send bug reports to bugs@pengutronix.de
 *
 * See COPYING for the status of this software.
 *
 * PURPOSE
 * This is a lightweight, single threaded multiconnection server for
 * data visualization based on JVisu applets.
 *
 * It's single threaded as it is intended to run on as much platforms
 * as possible. For uClinux a single threaded server is the best choice.
 * On standard linux platforms this concept does not hurt.
 *
 * All you have to provide is a structure with functions for symbol
 * resolving and variable write. Variable write can be NULL, get_symbol
 * must be a valid function.
 *
 * The server is running as long as the global variable jvisu_stop is zero.
 * You can stop the server by modifying jvisu_stop.
 * Signal handling is not done inside this code to make it as portable as
 * possible. You can do this in the thread which calls this code.
 *
 * Note that most communication related data is wrapped with special
 * datatypes which let you not access this data directly. They ensure
 * endianess conversion. Use the appropriate conversion macro.
 *
 */

#include <libpv_config.h>

#include <errno.h>
#include <fcntl.h>
#include <pthread.h>
#include <string.h>
#include <unistd.h>

#include <netinet/in.h>
#include <netinet/tcp.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>

#include <ptxlist.h>

#include <jv_types.h>
#include <pv_types.h>
#include <pv_lock.h>
#include <pv_bs.h>
#include <pv_register.h>

#define D1(x) if (_libpv_debuglevel > LEVEL_INFO) x

#define STAT_READ		1
#define STAT_WRITE		2

#define NSEC_PER_SEC    1000000000L

/**
 * set_normalized_timespec - set timespec sec and nsec parts and normalize
 *
 * @ts:		pointer to timespec variable to be set
 * @sec:	seconds to set
 * @nsec:	nanoseconds to set
 *
 * Set seconds and nanoseconds field of a timespec variable and
 * normalize to the timespec storage format
 *
 * Note: The tv_nsec part is always in the range of
 * 	0 <= tv_nsec < NSEC_PER_SEC
 * For negative values only the tv_sec field is negative !
 */
void set_normalized_timespec(struct timespec *ts, time_t sec, long nsec)
{
	while (nsec >= NSEC_PER_SEC) {
		nsec -= NSEC_PER_SEC;
		++sec;
	}
	while (nsec < 0) {
		nsec += NSEC_PER_SEC;
		--sec;
	}
	ts->tv_sec = sec;
	ts->tv_nsec = nsec;
}


/*
 * sub = lhs - rhs, in normalized form
 */
static inline struct timespec timespec_sub(struct timespec lhs,
					   struct timespec rhs)
{
	struct timespec ts_delta;
	set_normalized_timespec(&ts_delta, lhs.tv_sec - rhs.tv_sec,
				lhs.tv_nsec - rhs.tv_nsec);
	return ts_delta;
}


/**
 * timespec_to_ns - Convert timespec to nanoseconds
 * @ts:		pointer to the timespec variable to be converted
 *
 * Returns the scalar nanosecond representation of the timespec
 * parameter.
 */
static inline int64_t timespec_to_ns(const struct timespec *ts)
{
	return ((int64_t) ts->tv_sec * NSEC_PER_SEC) + ts->tv_nsec;
}


/* private prototypes */
static void unregister_vars (struct jv_server *jv_server, struct jvisu_sock *sd);
static void shutdown_sock (struct jv_server *jv_server, struct jvisu_sock *sd);
static void check_variables (struct jv_server *jv_server);
static int  do_accept (struct jv_server *jv_server, int sd);
static int  analyse_command (struct jv_server *jv_server, struct jvisu_sock *sd);
static int  process_request (struct jv_server *jv_server, struct jvisu_sock *sd);
static void check_fdset (struct jv_server *jv_server);
static int  get_symbol (struct jv_server *jv_server, struct jvisu_var *var, char *symbol);

static const char *prompt = "jvisu";

/** 
 *	Unregister all variables which are registered for a given 
 *	socket connection
 *  
 *	@param sd	socket description structure
 */
static void unregister_vars (struct jv_server *jv_server, struct jvisu_sock *sd)
{
	struct jvisu_var *var;
	int finished = 0; 

	if (is_empty(jv_server->jv_varlist))
		return;
	
	/* remove from list, free memory */
	for (goto_head(&jv_server->jv_varlist); !finished; goto_next(&jv_server->jv_varlist)) {
		
		var = (struct jvisu_var*)list_get_element_data(jv_server->jv_varlist);

		if (is_last(jv_server->jv_varlist))
			finished=1;
		
		if (var->sd->sd != sd->sd)
			continue;
		
		jv_server->jv_registeredvars--;
		jv_server->jv_varlist = list_del_element(jv_server->jv_varlist);
	}
}


/** 
 * 	Shut down a socket and unregister variables
 * 	
 *	@param sd	socket description structure
 */
static void shutdown_sock (struct jv_server *jv_server, struct jvisu_sock *sd)
{
	D1 (printf ("%s: %-20s: %d\n", prompt, __FUNCTION__, sd->sd));

	unregister_vars(jv_server, sd);
	jv_server->jv_connections--;
	FD_CLR (sd->sd, &jv_server->jv_read_fdset);
	FD_CLR (sd->sd, &jv_server->jv_write_fdset);
	close (sd->sd);
	sd->sd = -1;
}


/** 
 *	Retrieve a boolean variable from the application; do typecasting 
 *	from internal format
 *
 *	@param var	variable description structure
 *	@return		typecasted variable content 
 */
static inline u_char get_bool (struct jvisu_var *var)
{
	switch (var->inttype) {
		case jv_bool:
			return (*((u_char *) var->data) & var->bitmask) ? 1 : 0;
		case jv_short:
			return (*((uint16_t *) var->data)) ? 1 : 0;
		case jv_int:
			return (*((uint32_t *) var->data)) ? 1 : 0;
		case jv_float:
			return (*((float *) var->data)) ? 1 : 0;
		case jv_double:
			return (*((double *) var->data)) ? 1 : 0;
		case jv_string:	
		case jv_byte:
			return (*((u_char *) var->data)) ? 1 : 0;
		default:
			break;
	}	
	return 0;
}


/** 
 *	Retrieve a byte variable from the application; do typecasting 
 *	from internal format
 *	
 *	@param var	variable description structure
 *	@return		typecasted variable content 
 */
static inline u_char get_byte (struct jvisu_var *var)
{
	switch (var->inttype) {
		case jv_bool:
			return (*((u_char *) var->data) & var->bitmask) ? 1 : 0;
		case jv_short:
			return (u_char) (*((uint16_t *) var->data));
		case jv_int:
			return (u_char) (*((uint32_t *) var->data));
		case jv_float:
			return (u_char) (*((float *) var->data));
		case jv_double:
			return (u_char) (*((double *) var->data));
		case jv_string:	
		case jv_byte:
			return (*((u_char *) var->data));
		default:
			break;
	}	
	return 0;
}


/** 
 *	Retrieve a short variable from the application; do typecasting
 *	from internal format
 *	
 *	@param var	variable description structure
 *	@return		typecasted variable content 
 */
static inline short get_short (struct jvisu_var *var)
{
	switch (var->inttype) {
		case jv_bool:
			return (*((u_char *) var->data) & var->bitmask) ? 1 : 0;
		case jv_short:
			return (*((uint16_t *) var->data));
		case jv_int:
			return (short) (*((uint32_t *) var->data));
		case jv_float:
			return (short) (*((float *) var->data));
		case jv_double:
			return (short) (*((double *) var->data));
		case jv_string:	
			return (short) atoi ((char *) var->data);
		case jv_byte:
			return (short) (*((u_char *) var->data));
		default:
			break;
	}	
	return 0;
}


/** 
 *	Retrieve a integer variable from the application; do typecasting 
 *	from internal format
 *	
 *	@param var	variable description structure
 *	@return		typecasted variable content 
 */
static inline int get_int (struct jvisu_var *var)
{
	switch (var->inttype) {
		case jv_bool:
			return (*((u_char *) var->data) & var->bitmask) ? 1 : 0;
		case jv_short:
			return (int) (*((uint16_t *) var->data));
		case jv_int:
			return (*((uint32_t *) var->data));
		case jv_float:
			return (int) (*((float *) var->data));
		case jv_double:
			return (int) (*((double *) var->data));
		case jv_string:	
			return atoi ((char *) var->data);
		case jv_byte:
			return (int) (*((u_char *) var->data));
		default:
			break;
	}	
	return 0;
}


/** 
 *	Retrieve a float variable from the application; do typecasting 
 *	from internal format
 *
 *	@param var	variable description structure
 *	@return		typecasted variable content 
 */
static inline float get_float (struct jvisu_var *var)
{
	switch (var->inttype) {
		case jv_bool:
			return (*((u_char *) var->data) & var->bitmask) ? 1 : 0;
		case jv_short:
			return (float) (*((uint16_t *) var->data));
		case jv_int:
			return (float) (*((uint32_t *) var->data));
		case jv_float:
			return (*((float *) var->data));
		case jv_double:
			return (float) (*((double *) var->data));
		case jv_string:	
			return (float) atof ((char *) var->data);
		case jv_byte:
			return (float) (*((u_char *) var->data));
		default:
			break;
	}	
	return 0;
}


/** 
 *	Retrieve a double variable from the application; do typecasting 
 *	from internal format
 *	
 *	@param var	variable description structure
 *	@return		typecasted variable content 
 */
static inline double get_double (struct jvisu_var *var)
{
	switch (var->inttype) {
		case jv_bool:
			return (*((u_char *) var->data) & var->bitmask) ? 1 : 0;
		case jv_short:
			return (double) (*((uint16_t *) var->data));
		case jv_int:
			return (double) (*((uint32_t *) var->data));
		case jv_float:
			return (double) (*((float *) var->data));
		case jv_double:
			return (*((double *) var->data));
		case jv_string:	
			return atof ((char *) var->data);
		case jv_byte:
			return (double) (*((u_char *) var->data));
		default:
			break;
	}	
	return 0;
}


static inline int get_bytearray(struct jvisu_var *var, int8_t* dest)
{
	int i;
	int changed = 0;

	switch(var->inttype) {
		case jv_bytearray: {
			char *bp = (char *) var->data;
			for (i = 0; i < var->arraylen; i++) {
				if (bp[i] != var->cmp.bp[i]) {
					var->cmp.bp[i] = bp[i];	
					changed = 1;
				}
				dest[i] = bp[i];
			}		
			break;	
		}

		case jv_shortarray: {
			short *sp = (short *) var->data;
			for (i = 0; i < var->arraylen; i++) {
				if (sp[i] != var->cmp.sp[i]) {
					var->cmp.sp[i] = sp[i];	
					changed = 1;
				}
				dest[i] = (int8_t)sp[i];
			}		
			break;	
		}

		case jv_intarray: {
			int *ip = (int *) var->data;
			for (i = 0; i < var->arraylen; i++) {
				if (ip[i] != var->cmp.ip[i]) {
					var->cmp.ip[i] = ip[i];	
					changed = 1;
				}
				dest[i] = (int8_t)ip[i];
			}		
			break;	
		}

		case jv_floatarray: {
			float *fp = (float *) var->data;
			for (i = 0; i < var->arraylen; i++) {
				if (fp[i] != var->cmp.fp[i]) {
					var->cmp.fp[i] = fp[i];	
					changed = 1;
				}
				dest[i] = (int8_t)fp[i];
			}		
			break;	
		}

		case jv_doublearray: {
			double *dp = (double *) var->data;
			for (i = 0; i < var->arraylen; i++) {
				if (dp[i] != var->cmp.dp[i]) {
					var->cmp.dp[i] = dp[i];	
					changed = 1;
				}
				dest[i] = (int8_t)dp[i];
			}		
			break;	
		}

		default: {
			break;
		}
	}

	return changed;
}


static inline int get_shortarray(struct jvisu_var *var, int16_t* dest)
{
	int i;
	int changed = 0;


	switch(var->inttype) {
		case jv_bytearray: {
			char *bp = (char *) var->data;
			for (i = 0; i < var->arraylen; i++) {
				if (bp[i] != var->cmp.bp[i]) {
					var->cmp.bp[i] = bp[i];	
					changed = 1;
				}
				dest[i] = cpu_to_je16( (int16_t)bp[i] ).v16;
			}		
			break;	
		}

		case jv_shortarray: {
			short *sp = (short *) var->data;
			for (i = 0; i < var->arraylen; i++) {
				if (sp[i] != var->cmp.sp[i]) {
					var->cmp.sp[i] = sp[i];	
					changed = 1;
				}
				dest[i] = cpu_to_je16( (int16_t)sp[i] ).v16;
			}		
			break;	
		}

		case jv_intarray: {
			int *ip = (int *) var->data;
			for (i = 0; i < var->arraylen; i++) {
				if (ip[i] != var->cmp.ip[i]) {
					var->cmp.ip[i] = ip[i];	
					changed = 1;
				}
				dest[i] = cpu_to_je16( (int16_t)ip[i] ).v16;
			}		
			break;	
		}

		case jv_floatarray: {
			float *fp = (float *) var->data;
			for (i = 0; i < var->arraylen; i++) {
				if (fp[i] != var->cmp.fp[i]) {
					var->cmp.fp[i] = fp[i];	
					changed = 1;
				}
				dest[i] = cpu_to_je16( (int16_t)fp[i] ).v16;
			}		
			break;	
		}

		case jv_doublearray: {
			double *dp = (double *) var->data;
			for (i = 0; i < var->arraylen; i++) {
				if (dp[i] != var->cmp.dp[i]) {
					var->cmp.dp[i] = dp[i];	
					changed = 1;
				}
				dest[i] = cpu_to_je16( (int16_t)dp[i] ).v16;
			}		
			break;	
		}

		default: {
			break;
		}
	}

	return changed;
}

static inline int get_intarray(struct jvisu_var *var, int32_t* dest)
{
	int i;
	int changed = 0;

	switch(var->inttype) {
		case jv_bytearray: {
			char *bp = (char *) var->data;
			for (i = 0; i < var->arraylen; i++) {
				if (bp[i] != var->cmp.bp[i]) {
					var->cmp.bp[i] = bp[i];	
					changed = 1;
				}
				dest[i] = cpu_to_je32( (int32_t)bp[i] ).v32;
			}		
			break;	
		}

		case jv_shortarray: {
			short *sp = (short *) var->data;
			for (i = 0; i < var->arraylen; i++) {
				if (sp[i] != var->cmp.sp[i]) {
					var->cmp.sp[i] = sp[i];	
					changed = 1;
				}
				dest[i] = cpu_to_je32( (int32_t)sp[i] ).v32;
			}		
			break;	
		}

		case jv_intarray: {
			int *ip = (int *) var->data;
			for (i = 0; i < var->arraylen; i++) {
				if (ip[i] != var->cmp.ip[i]) {
					var->cmp.ip[i] = ip[i];	
					changed = 1;
				}
				dest[i] = cpu_to_je32( (int32_t)ip[i] ).v32;
			}		
			break;	
		}

		case jv_floatarray: {
			float *fp = (float *) var->data;
			for (i = 0; i < var->arraylen; i++) {
				if (fp[i] != var->cmp.fp[i]) {
					var->cmp.fp[i] = fp[i];	
					changed = 1;
				}
				dest[i] = cpu_to_je32( (int32_t)fp[i] ).v32;
			}		
			break;	
		}

		case jv_doublearray: {
			double *dp = (double *) var->data;
			for (i = 0; i < var->arraylen; i++) {
				if (dp[i] != var->cmp.dp[i]) {
					var->cmp.dp[i] = dp[i];	
					changed = 1;
				}
				dest[i] = cpu_to_je32( (int32_t)dp[i] ).v32;
			}		
			break;	
		}

		default: {
			break;
		}
	}

	return changed;
}

static inline int get_floatarray(struct jvisu_var *var, float* dest)
{
	int i;
	int changed = 0;

	switch(var->inttype) {
		case jv_bytearray: {
			char *bp = (char *) var->data;
			for (i = 0; i < var->arraylen; i++) {
				if (bp[i] != var->cmp.bp[i]) {
					var->cmp.bp[i] = bp[i];	
					changed = 1;
				}
				dest[i] = cpu_to_je32( (float)bp[i] ).v32;
			}		
			break;	
		}

		case jv_shortarray: {
			short *sp = (short *) var->data;
			for (i = 0; i < var->arraylen; i++) {
				if (sp[i] != var->cmp.sp[i]) {
					var->cmp.sp[i] = sp[i];	
					changed = 1;
				}
				dest[i] = cpu_to_je32( (float)sp[i] ).v32;
			}		
			break;	
		}

		case jv_intarray: {
			int *ip = (int *) var->data;
			for (i = 0; i < var->arraylen; i++) {
				if (ip[i] != var->cmp.ip[i]) {
					var->cmp.ip[i] = ip[i];	
					changed = 1;
				}
				dest[i] = cpu_to_je32( (float)ip[i] ).v32;
			}		
			break;	
		}

		case jv_floatarray: {
			float *fp = (float *) var->data;
			for (i = 0; i < var->arraylen; i++) {
				if (fp[i] != var->cmp.fp[i]) {
					var->cmp.fp[i] = fp[i];	
					changed = 1;
				}
				dest[i] = cpu_to_je32( (float)fp[i] ).v32;
			}		
			break;	
		}

		case jv_doublearray: {
			double *dp = (double *) var->data;
			for (i = 0; i < var->arraylen; i++) {
				if (dp[i] != var->cmp.dp[i]) {
					var->cmp.dp[i] = dp[i];	
					changed = 1;
				}
				dest[i] = cpu_to_je32( (float)dp[i] ).v32;
			}		
			break;	
		}

		default: {
			break;
		}
	}

	return changed;
}

static inline int get_doublearray(struct jvisu_var *var, double* dest)
{
	int i;
	int changed = 0;

	switch(var->inttype) {
		case jv_bytearray: {
			char *bp = (char *) var->data;
			for (i = 0; i < var->arraylen; i++) {
				if (bp[i] != var->cmp.bp[i]) {
					var->cmp.bp[i] = bp[i];	
					changed = 1;
				}
				dest[i] = cpu_to_je64( (double)bp[i] ).v64;
			}		
			break;	
		}

		case jv_shortarray: {
			short *sp = (short *) var->data;
			for (i = 0; i < var->arraylen; i++) {
				if (sp[i] != var->cmp.sp[i]) {
					var->cmp.sp[i] = sp[i];	
					changed = 1;
				}
				dest[i] = cpu_to_je64( (double)sp[i] ).v64;
			}		
			break;	
		}

		case jv_intarray: {
			int *ip = (int *) var->data;
			for (i = 0; i < var->arraylen; i++) {
				if (ip[i] != var->cmp.ip[i]) {
					var->cmp.ip[i] = ip[i];	
					changed = 1;
				}
				dest[i] = cpu_to_je64( (double)ip[i] ).v64;
			}		
			break;	
		}

		case jv_floatarray: {
			float *fp = (float *) var->data;
			for (i = 0; i < var->arraylen; i++) {
				if (fp[i] != var->cmp.fp[i]) {
					var->cmp.fp[i] = fp[i];	
					changed = 1;
				}
				dest[i] = cpu_to_je64( (double)fp[i] ).v64;
			}		
			break;	
		}

		case jv_doublearray: {
			double *dp = (double *) var->data;
			for (i = 0; i < var->arraylen; i++) {
				if (dp[i] != var->cmp.dp[i]) {
					var->cmp.dp[i] = dp[i];	
					changed = 1;
				}
				dest[i] = cpu_to_je64( (double)dp[i] ).v64;
			}		
			break;	
		}

		default: {
			break;
		}
	}

	return changed;
}


/** 
 *	Check if variables have been added or changed
 */
static void check_variables (struct jv_server *jv_server)
{
	struct timespec		tnow, tdiff;
	int			ms;
	int			finished = 0;

	if (is_empty(jv_server->jv_varlist))
		return; 

	/* read current time */
	clock_gettime(CLOCK_MONOTONIC, &tnow);
	tdiff = timespec_sub(tnow, jv_server->jv_tact);

	/* convert to ms */
	ms = timespec_to_ns(&tdiff) / 1000 / 1000;
	if (ms <= 0) {
		return;
	}

	/* save current time */
	jv_server->jv_tact = tnow;

	/* walk through the list */
	for (goto_head(&jv_server->jv_varlist); !finished; goto_next(&jv_server->jv_varlist)) {
		
		size_t			len;
		int 			res, force;
		struct jvisu_sndvar	*snd;
		struct jvisu_var 	*var = (struct jvisu_var*)list_get_element_data(jv_server->jv_varlist);
		var->elapsed += ms;

		if (is_last(jv_server->jv_varlist))
			finished=1;
		
		/* time elapsed ? */
		if (var->elapsed < var->rate)
			continue;

		/* transfer in progress ? */
		if (var->sd->status || (var->permissions & JVISU_PERM_WO)) 
			continue;

		snd = (struct jvisu_sndvar *) var->sd->buf;
		snd->hdr.cmd = cmd_send_var;
		snd->hdr.type = (u_char) var->type;
		len = sizeof (struct jvisu_sndvar);
		snd->idx = cpu_to_je32 (var->idx);
	
		force = (var->status & JVISU_STAT_INIT) ? 0 : 1;
		
		switch (var->type) {
		case jv_byte: {
			pv_lock_lock(jv_server->pv_engine);
			u_char act = get_byte (var);
			pv_lock_unlock(jv_server->pv_engine);
			if ((var->cmp.b) == act && !force)
				continue;
			var->cmp.b = act;
			len++;
			snd->data.b[0] = act;
			break;
		}	
		case jv_short: {
			pv_lock_lock(jv_server->pv_engine);
			short act = get_short (var);
			pv_lock_unlock(jv_server->pv_engine);
			if ((var->cmp.s) == act && !force)
				continue;
			var->cmp.s = act;
			len += sizeof (short);
			snd->data.s[0] = cpu_to_je16 (act);
			break;
		}	
		case jv_int: {
			pv_lock_lock(jv_server->pv_engine);
			int act = get_int (var);
			pv_lock_unlock(jv_server->pv_engine);
			if ((var->cmp.i) == act && !force)
				continue;
			var->cmp.i = act;
			len += sizeof (int);
			snd->data.i[0] = cpu_to_je32 (act);
			break;
		}	
		case jv_float: {
			union { float f; uint32_t i; } act;
			pv_lock_lock(jv_server->pv_engine);
			act.f = get_float (var);
			pv_lock_unlock(jv_server->pv_engine);
			if ((var->cmp.f) == act.f && !force)
				continue;
			var->cmp.f = act.f;
			len += sizeof (float);
			snd->data.i[0] = cpu_to_je32 (act.i);
			break;
		}	
		case jv_double: {
			union { double d; uint64_t i; } act;
			pv_lock_lock(jv_server->pv_engine);
			act.d = get_double (var);
			pv_lock_unlock(jv_server->pv_engine);
			if ((var->cmp.d) == act.d && !force)
				continue;
			var->cmp.d = act.d;
			len += sizeof (double);
			snd->data.d[0] = cpu_to_je64 (act.i);
			break;
		}	
		case jv_bool: {
			pv_lock_lock(jv_server->pv_engine);
			u_char act = get_bool (var);
			pv_lock_unlock(jv_server->pv_engine);
			if ((var->cmp.b == act) && !force)
				continue;
			var->cmp.b = act;
			len++;
			snd->data.b[0] = act;
			break;
		}
			      
		/* 
		 * TYPE String provides no typecasting ATM
		 * As I have no real good idea to track changes on strings
		 * due to variable length and content I decided to transfer strings 
		 * in the given interval. 
		 * If you want to have change triggered string display 
		 * then use a index variable and provide the strings in a file
		 * which can be retrieved by the applet 
		 * If you are still not happy provide a better solution :)
		 */

		case jv_string:
			len += strlen ((char *)var->data);
			strcpy ((char*)snd->data.b, (char *)var->data);
			break;

		/* 
		 * TYPE Stream is a special case. Check, if the length variable
		 * is > 0. Send out the bytes and set it to zero
		 */
			
		case jv_stream: {
			struct jvisu_varstream *p = (struct jvisu_varstream*) var->data;
			if (!p->len)
				continue;
			len += p->len;
			D1 (
				printf ("%s: %-20s: jv_stream: Send %d bytes (total %zd)\n", 
				prompt, __FUNCTION__, p->len, len)
			);
			memcpy (snd->data.b, p->data, (size_t)p->len);
			p->len = 0;
			break;
		}

		case jv_bytearray: {
			pv_lock_lock(jv_server->pv_engine);

			if ( get_bytearray(var, (int8_t*)snd->data.b ) == 1)
				force = 1;

			pv_lock_unlock(jv_server->pv_engine);

			if (!force)
				continue;

			len += var->arraylen * sizeof (int8_t);
			break;
		}
		case jv_shortarray: {
			pv_lock_lock(jv_server->pv_engine);

			if ( get_shortarray(var, (int16_t*)snd->data.s ) == 1)
				force = 1;

			pv_lock_unlock(jv_server->pv_engine);

			if (!force)
				continue;

			len += var->arraylen * sizeof (int16_t);
			break;
		}
		case jv_intarray: {
			pv_lock_lock(jv_server->pv_engine);

			if ( get_intarray(var, (int32_t*)snd->data.i ) == 1)
				force = 1;

			pv_lock_unlock(jv_server->pv_engine);

			if (!force)
				continue;

			len += var->arraylen * sizeof (int32_t);
			break;
		}
		case jv_floatarray: {
			pv_lock_lock(jv_server->pv_engine);

			if ( get_floatarray(var, (float*)snd->data.i ) == 1)
				force = 1;

			pv_lock_unlock(jv_server->pv_engine);

			if (!force)
				continue;

			len += var->arraylen * sizeof (float);
			break;

		}
		case jv_doublearray: {
			pv_lock_lock(jv_server->pv_engine);

			if ( get_doublearray(var, (double*)snd->data.d ) == 1)
				force = 1;

			pv_lock_unlock(jv_server->pv_engine);

			if (!force)
				continue;

			len += var->arraylen * sizeof (double);
			break;
		}
		default:
			continue;
		}

		/* send value */
		snd->hdr.len = cpu_to_je16 ((ushort) len);
		res = write (var->sd->sd, var->sd->buf, len);
		if (res < 0) {
			if (res != EAGAIN) {
				shutdown_sock(jv_server, var->sd);
				continue;
			}
			res = 0;
		}
		/* all sent ? */
		if (res == len) {
			var->elapsed = 0;	
			var->status |= JVISU_STAT_INIT;
			continue;
		}
		fprintf (stderr, "%s: %-20s: FIXME: Send has to be delayed\n", prompt, __FUNCTION__);	
		var->sd->bufpoi = res;
		var->sd->cmdlen = len;
		var->sd->status	= STAT_WRITE;

	}	
}


/** 
 *	Write a boolean variable in the application; do typecasting 
 *	from applet format
 *	
 *	@param var	variable description structure
 *	@param dat	union containing the data to write
 */
static inline void set_bool (struct jvisu_var *var, union jvisu_pv *pv)
{
	int res;

	switch (var->type) {
		case jv_bool:
		case jv_string:	
		case jv_byte:
			res = pv->b[0] ? 1 : 0;
			break;
		case jv_short:
			res = pv->s[0] ? 1 : 0;
			break;
		case jv_int:
			res = pv->i[0] ? 1 : 0;
			break;
		case jv_float:
			res = (pv->f[0] != 0.0) ? 1 : 0;
			break;
		case jv_double:
			res = (pv->d[0] != 0.0) ? 1 : 0;
			break;
		default:
			return;
	}
	if (res)
		*((u_char *) var->data) |= var->bitmask;
	else	
		*((u_char *) var->data) &= ~var->bitmask;
}


/** 
 *	Write a byte variable in the application; do typecasting 
 *	from applet format
 *	
 *	@param var	variable description structure
 *	@param dat	union containing the data to write
 */
static inline void set_byte (struct jvisu_var *var, union jvisu_pv *pv)
{

	switch (var->type) {
		case jv_bool:
		case jv_byte:
			*((u_char *) var->data) = pv->b[0];
			break;
		case jv_short:
			*((u_char *) var->data) = (u_char) pv->s[0];
			break;
		case jv_int:
			*((u_char *) var->data) = (u_char) pv->i[0];
			break;
		case jv_float:
			*((u_char *) var->data) = (u_char) pv->f[0];
			break;
		case jv_double:
			*((u_char *) var->data) = (u_char) pv->d[0];
			break;
		case jv_string:
			*((u_char *) var->data) = (u_char) atoi ((char*)pv->b);
			break;
		default:
			return;
	}
}


/** 
 *	Write a short variable in the application; do typecasting 
 *	from applet format
 *	
 *	@param var	variable description structure
 *	@param dat	union containing the data to write
 */
static inline void set_short (struct jvisu_var *var, union jvisu_pv *pv)
{

	switch (var->type) {
		case jv_bool:
		case jv_byte:
			*((short *) var->data) = (short) pv->b[0];
			break;
		case jv_short:
			*((short *) var->data) = pv->s[0];
			break;
		case jv_int:
			*((short *) var->data) = (short) pv->i[0];
			break;
		case jv_float:
			*((short *) var->data) = (short) pv->f[0];
			break;
		case jv_double:
			*((short *) var->data) = (short) pv->d[0];
			break;
		case jv_string:
			*((short *) var->data) = (short) atoi ((char*)pv->b);
			break;
		default:
			return;
	}
}


/** 
 *	Write a integer variable in the application; do typecasting 
 *	from applet format
 *	
 *	@param var	variable description structure
 *	@param pv	union containing the data to write
 */
static inline void set_int (struct jvisu_var *var, union jvisu_pv *pv)
{

	switch (var->type) {
		case jv_bool:
		case jv_byte:
			D1 (
				printf ("%s: %-20s: var data: %d\n", 
				prompt, __FUNCTION__, (int) pv->b[0])
			);
			*((int *) var->data) = (int) pv->b[0];
			break;
		case jv_short:
			*((int *) var->data) = (int) pv->s[0];
			break;
		case jv_int:
			*((int *) var->data) = pv->i[0];
			break;
		case jv_float:
			*((int *) var->data) = (int) pv->f[0];
			break;
		case jv_double:
			*((int *) var->data) = (int) pv->d[0];
			break;
		case jv_string:
			*((int *) var->data) = atoi ((char*)pv->b);
			break;
		default:
			return;
	}
}


/** 
 *	Write a float variable in the application; do typecasting 
 *	from applet format
 *	
 *	@param var	variable description structure
 *	@param pv	union containing the data to write
 */
static inline void set_float (struct jvisu_var *var, union jvisu_pv *pv)
{

	switch (var->type) {
		case jv_bool:
		case jv_byte:
			*((float *) var->data) = (float) pv->b[0];
			break;
		case jv_short:
			*((float *) var->data) = (float) pv->s[0];
			break;
		case jv_int:
			*((float *) var->data) = (float) pv->i[0];
			break;
		case jv_float:
			*((float *) var->data) = pv->f[0];
			break;
		case jv_double:
			*((float *) var->data) = (float) pv->d[0];
			break;
		case jv_string:
			*((float *) var->data) = (float) atof ((char*)pv->b);
			break;
		default:
			return;
	}
}


/** 
 *	Write a double variable in the application; do typecasting 
 *	from applet format
 *	
 *	@param var	variable description structure
 *	@param pv	union containing the data to write
 */
static inline void set_double (struct jvisu_var *var, union jvisu_pv *pv)
{

	switch (var->type) {
		case jv_bool:
		case jv_byte:
			*((double *) var->data) = (double) pv->b[0];
			break;
		case jv_short:
			*((double *) var->data) = (double) pv->s[0];
			break;
		case jv_int:
			*((double *) var->data) = (double) pv->i[0];
			break;
		case jv_float:
			*((double *) var->data) = (double) pv->f[0];
			break;
		case jv_double:
			*((double *) var->data) = pv->d[0];
			break;
		case jv_string:
			*((double *) var->data) = atof ((char*)pv->b);
			break;
		default:
			return;
	}
}


/** 
 *	Write a string variable in the application; do typecasting 
 *	from applet format
 *	
 *	@param var	variable description structure
 *	@param pv	union containing the data to write
 *      @return		the number of bytes copied
 */
static inline int set_string (struct jvisu_var *var, union jvisu_pv *pv)
{
	int ret;

	switch (var->type) {
	case jv_bool:
	case jv_byte:
		ret = sprintf ((char *) var->data, "%d", (int) pv->b[0]);
		break;
	case jv_short:
		ret = sprintf ((char *) var->data, "%d", (int) pv->s[0]);
		break;
	case jv_int:
		ret = sprintf ((char *) var->data, "%d", pv->i[0]);
		break;
	case jv_float:
		ret = sprintf ((char *) var->data, "%f", pv->f[0]);
		break;
	case jv_double:
		ret = sprintf ((char *) var->data, "%f", pv->f[0]);
		break;
	case jv_string:
		ret = strlen((char*)pv->b) + 1;
		strcpy ((char*)var->data, (char*)pv->b);
		break;
	default:
		return -1;
	}
	return ret;
}




/** 
 *	Find a matching variable and endianess-convert the incoming data, 
 *	then call the appropriate write function depending on internal
 *	variable type; call the variable specific write function if available
 *
 *	@param pv	receive buffer structure
 *	@param sd	socket descriptor on which the write command was received
 */
static void write_variable (struct jv_server *jv_server, struct jvisu_sndvar	*dat, struct jvisu_sock *sd)
{
	char			buf[MAX_DATA_LEN];
	union jvisu_pv	 	*idat = (union jvisu_pv *) buf;
	int 			n, finished = 0; 
	
	D1(
		printf ("%s: %-20s: idx = %d\n", prompt, __FUNCTION__, je32_to_cpu(dat->idx))
	);	

	if (is_empty(jv_server->jv_varlist)) {
		D1(printf("%s: %-20s: no registered variables\n", prompt, __FUNCTION__));
		return;
	}
	
	/* find the matching entry */	
	for (goto_head(&jv_server->jv_varlist); !finished; goto_next(&jv_server->jv_varlist)) {

		struct jvisu_var *var = (struct jvisu_var*)list_get_element_data(jv_server->jv_varlist);

		if (is_last(jv_server->jv_varlist))
			finished=1;
		
		if (var->idx != je32_to_cpu (dat->idx) || var->sd != sd)
			continue;
			
		/* check, if this variable is allowed to be written */
		if ((var->permissions & JVISU_PERM_RO)) {
			D1 (printf ("%s: %-20s: request to write readonly var!\n", prompt, __FUNCTION__));
			return;
		}	
		D1 (printf ("%s: %-20s: type: %d, %d\n", prompt, __FUNCTION__, var->type, var->inttype));	
		
		/* do endianess conversion */	
		switch (var->type) {
			case jv_byte:
			case jv_bool:
				idat->b[0] = dat->data.b[0];
				break;
			case jv_short:
				idat->u16[0] = je16_to_cpu (dat->data.s[0]);
				break;
			case jv_int:
			case jv_float:
				idat->u32[0] = je32_to_cpu (dat->data.i[0]);
				break;
			case jv_double: {
				idat->u64[0] = je64_to_cpu (dat->data.d[0]);
				break;
			}	
			case jv_string:
				dat->data.b[(sd->cmdlen - sizeof (struct jvisu_cmdhdr))] = '\0';
				strcpy ((char*)idat->b, (char*)dat->data.b);
				break;
			case jv_stream:
				break;	
			case jv_bytearray: {
				int i;
				for (i = 0; i < var->arraylen; i++) {
					idat->b[i] = dat->data.b[i];
				}			
				break;	
			}
			case jv_shortarray: {
				int i;
				for (i = 0; i < var->arraylen; i++) {
					idat->u16[i] = je16_to_cpu (dat->data.s[i]);
				}
				break;	
			}
			case jv_intarray: {
				int i;
				for (i = 0; i < var->arraylen; i++) {
					idat->u32[i] = je32_to_cpu (dat->data.i[i]);
				}
				break;	
			}
			case jv_floatarray: {
				int i;
				for (i = 0; i < var->arraylen; i++) {
					idat->f[i] = je32_to_cpu (dat->data.i[i]);
				}
				break;	
			}
			case jv_doublearray: {
				int i;
				for (i = 0; i < var->arraylen; i++) {
					idat->d[i] = je64_to_cpu (dat->data.d[i]);
				}
				break;	
			}
			default:
				return;	
		}

		/* call the proper write function*/	
		switch (var->inttype) {
			case jv_bool:
				pv_lock_lock(jv_server->pv_engine);
				set_bool (var, idat); 
				pv_bs_flush_abs(jv_server->pv_engine, var->data, 1);
				pv_lock_unlock(jv_server->pv_engine);
				break;;
			case jv_byte:
				pv_lock_lock(jv_server->pv_engine);
				set_byte (var, idat); 
				pv_bs_flush_abs(jv_server->pv_engine, var->data, 1);
				pv_lock_unlock(jv_server->pv_engine);
				break;
			case jv_short:
				pv_lock_lock(jv_server->pv_engine);
				set_short (var, idat); 
				pv_bs_flush_abs(jv_server->pv_engine, var->data, sizeof(short));
				pv_lock_unlock(jv_server->pv_engine);
				break;
			case jv_int:
				pv_lock_lock(jv_server->pv_engine);
				set_int (var, idat); 
				pv_bs_flush_abs(jv_server->pv_engine, var->data, sizeof(int));
				pv_lock_unlock(jv_server->pv_engine);
				break;
			case jv_float:
				pv_lock_lock(jv_server->pv_engine);
				set_float (var, idat); 
				pv_bs_flush_abs(jv_server->pv_engine, var->data, sizeof(float));
				pv_lock_unlock(jv_server->pv_engine);
				break;
			case jv_double:
				pv_lock_lock(jv_server->pv_engine);
				set_double (var, idat);
				pv_bs_flush_abs(jv_server->pv_engine, var->data, sizeof(double));
				pv_lock_unlock(jv_server->pv_engine);
				break;
			case jv_string:
				pv_lock_lock(jv_server->pv_engine);
				n = set_string (var, idat); 
				pv_bs_flush_abs(jv_server->pv_engine, var->data, n);
				pv_lock_unlock(jv_server->pv_engine);
				break;
			case jv_stream: {
				struct jvisu_varstream *p = (struct jvisu_varstream*) var->data;
				pv_lock_lock(jv_server->pv_engine);
				memcpy (p->data, dat->data.b, (sd->cmdlen - sizeof (struct jvisu_cmdhdr)));
				/* FIXME: flush (mkl) */
				pv_lock_unlock(jv_server->pv_engine);
				p->len = (sd->cmdlen - sizeof (struct jvisu_cmdhdr));
				break;
				}	
			case jv_bytearray: {
				/* check length */
				int num = sd->cmdlen - sizeof (struct jvisu_cmdhdr);
				if (num > var->arraylen)
					num = var->arraylen;
				pv_lock_lock(jv_server->pv_engine);
				memcpy (var->data, dat->data.b, (size_t) num);
				pv_bs_flush_abs(jv_server->pv_engine, var->data, (size_t) num);
				pv_lock_unlock(jv_server->pv_engine);
				D1 (printf ("%s: %-20s: jv_bytearray: len = %d\n", prompt, __FUNCTION__, num));
				break;
				}
			case jv_shortarray: {
				/* check length */
				int num = (sd->cmdlen - sizeof (struct jvisu_cmdhdr)) / sizeof(uint16_t);
				if (num > var->arraylen)
					num = var->arraylen;
				pv_lock_lock(jv_server->pv_engine);
				memcpy (var->data, dat->data.b, (size_t) num * sizeof(uint16_t));
				pv_bs_flush_abs(jv_server->pv_engine, var->data, (size_t) num * sizeof(uint16_t));
				pv_lock_unlock(jv_server->pv_engine);
				D1 (printf ("%s: %-20s: jv_shortarray: len = %d\n", prompt, __FUNCTION__, num));
				break;
				}
			case jv_intarray: {
				/* check length */
				int num = (sd->cmdlen - sizeof (struct jvisu_cmdhdr)) / sizeof(uint32_t);
				if (num > var->arraylen)
					num = var->arraylen;
				pv_lock_lock(jv_server->pv_engine);
				memcpy (var->data, dat->data.b, (size_t) num * sizeof(uint32_t));
				pv_bs_flush_abs(jv_server->pv_engine, var->data, (size_t) num * sizeof(uint32_t));
				pv_lock_unlock(jv_server->pv_engine);
				D1 (printf ("%s: %-20s: jv_intarray: len = %d\n", prompt, __FUNCTION__, num));
				break;
				}
			case jv_floatarray: {
				/* check length */
				int num = (sd->cmdlen - sizeof (struct jvisu_cmdhdr)) / sizeof(float);
				if (num > var->arraylen)
					num = var->arraylen;
				pv_lock_lock(jv_server->pv_engine);
				memcpy (var->data, dat->data.b, (size_t) num * sizeof(float));
				pv_bs_flush_abs(jv_server->pv_engine, var->data, (size_t) num * sizeof(float));
				pv_lock_unlock(jv_server->pv_engine);
				D1 (printf ("%s: %-20s: jv_floatarray: len = %d\n", prompt, __FUNCTION__, num));
				break;
				}
			case jv_doublearray: {
				/* check length */
				int num = (sd->cmdlen - sizeof (struct jvisu_cmdhdr)) / sizeof(double);
				if (num > var->arraylen)
					num = var->arraylen;
				pv_lock_lock(jv_server->pv_engine);
				memcpy (var->data, dat->data.b, (size_t) num * sizeof(double));
				pv_bs_flush_abs(jv_server->pv_engine, var->data, (size_t) num * sizeof(double));
				pv_lock_unlock(jv_server->pv_engine);
				D1 (printf ("%s: %-20s: jv_doublearray: len = %d\n", prompt, __FUNCTION__, num));
				break;
				}
		}
		
		/* check if a write_done function for this variable is available */
// RSC
//		if (var->write_done)
//			var->write_done (var->arg);
		
		var->elapsed += var->rate;	
	}
}


/** 
 *	Start socket
 *	
 *	@param port	portnumber to listen
 *	@return		filedescriptor or errorcode
 */
int jv_server_start_socket (struct jv_server *jv_server, int port)
{
	int			sd, opt;
	struct sockaddr_in 	sin;	

	/* open the socket */
	sd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (sd < 0)
		return sd;
		
	/* set non blocking */
	if (fcntl(sd, F_SETFL, O_NONBLOCK) < 0) 
		goto err;

	/* close on exit */
	if (fcntl(sd, F_SETFD, 1) < 0)
		goto err;
	
	/* reuse address */
	opt = 1;
	if (setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, (void *) &opt, sizeof(opt)) < 0)
		goto err;

	/* fill in the neccecary stuff */
	memset(&sin, 0, sizeof(sin));
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = htonl (INADDR_ANY);
	sin.sin_port = htons ((uint16_t) port);		

	/* bind it */
	if (bind (sd, (struct sockaddr *) &sin, sizeof (sin)) < 0)
		goto err;
	
	/* try to listen with max connections */
	if (listen(sd, MAX_CONNECTIONS) < 0)
		goto err;

	return sd;
	
err:
	fprintf (stderr, "%s: %-20s: error\n", prompt, __FUNCTION__);
	close (sd);
	return -1;
}


/** 
 *	Accept connections if we have less than MAX_CONNECTIONS
 *	
 *	@param sd	socket descriptor
 *	@return		new socket descriptor or error
 */
static int do_accept (struct jv_server *jv_server, int sd)
{
	int 			i, fd, opt = 1;
	struct sockaddr_in 	raddr;
	socklen_t		addrlen = sizeof (raddr);

	D1 (printf ("%s: %-20s: accepting connection\n", prompt, __FUNCTION__));

	/* sorry, enough connections */
	if (jv_server->jv_connections > MAX_CONNECTIONS) {
		D1(
			printf ("%s: %-20s: error: more than MAX_CONNECTIONS open\n", prompt, __FUNCTION__)
		);
		return 0;
	}

	/* try to accept the connection */
  	raddr.sin_family = 0xdead;
	fd = accept(sd, (struct sockaddr *) &raddr, &addrlen);

	if (fd < 0) {
		D1 (printf ("%s: %-20s: error: failed!\n", prompt, __FUNCTION__));
		if (errno == EAGAIN || errno == EWOULDBLOCK)
			return 0;
		else 	
			return fd;
	}

	/* some linux kernels have this problem when client closes between
	 * select() and accept()	
	 */
	if (raddr.sin_family != AF_INET)
		goto err;

	/* keep alive */
	if (setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE, (void *) &opt, sizeof(opt)) < 0)
		goto err;
	/* reuse address */
	if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (void *) &opt, sizeof(opt)) < 0)
		goto err;
	/* no delay */
	if (setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, (void *) &opt, sizeof (opt)) < 0)
		goto err;
	
	/* non blocking mode */
	if (fcntl (fd, F_SETFL, O_NONBLOCK) < 0) 
		goto err;

	/* close on exit */
	if (fcntl(fd, F_SETFD, 1) < 0)
		goto err;

	/* put it into an empty place (-1 => empty) */
	for (i = 0; i <= MAX_CONNECTIONS; i++) {
		if (jv_server->jv_active_sock[i].sd == -1) {
			jv_server->jv_active_sock[i].sd = fd;
			jv_server->jv_active_sock[i].status = 0;
			jv_server->jv_active_sock[i].bufpoi = 0;
			jv_server->jv_active_sock[i].cmdlen = 0;
			jv_server->jv_connections++;
			D1(
				printf ("%s: %-20s: socket descriptor stored: sd=%d, currently %d connection%s active\n", 
				prompt, __FUNCTION__, fd, jv_server->jv_connections, jv_server->jv_connections==1 ? "" : "s")
			);
			return 0;	
		}
	}
	/* should not happen */
	jv_server->jv_connections = MAX_CONNECTIONS;
	fprintf (stderr, "%s: %-20s: eeep! no free activ_sock, but connections < max\n", prompt, __FUNCTION__);
		
	/* error exit for valid socket descriptors */
err:
	fprintf (stderr, "%s: %-20s: error occured, exiting\n", prompt, __FUNCTION__);
	close(fd);
	return 0;
}


/** 
 *	Analyse command received via JVisu Native Protocol
 *	
 *	@param sd	jvisu socket description structure
 *	@return		0 on success, else error number
 */
static int analyse_command (struct jv_server *jv_server, struct jvisu_sock *sd)
{
	int	res = 0;

	switch (sd->buf[0]) {
	
	/* 
	 * the applets wants to register a variable with us
	 * FIXME: handle string variables
	 */

	case cmd_register_var: {

		struct jvisu_var var;
		struct jvisu_regvar_cmd	*reg_cmd = (struct jvisu_regvar_cmd *) sd->buf;
		int namelen;

		/* truncate symbol, if neccecary */
		namelen = MIN (sd->bufpoi - sizeof (struct jvisu_regvar_cmd), MAX_SYMBOL_LEN);

		/* terminate string */
		sd->buf[sd->cmdlen] = 0x0;

		/* copy data from network command into jvisu_var */
		var.sd = sd;
		var.type = (int)reg_cmd->hdr.type;
		var.idx = je32_to_cpu(reg_cmd->idx);
		var.rate = je32_to_cpu(reg_cmd->rate);
		var.elapsed = var.rate;
		
		D1(
			printf ("%s: %-20s: cmd_register_var(%s), idx=%d, type=%d\n", 
			prompt, __FUNCTION__, reg_cmd->symbol, var.idx, var.type)
		);

		/* get a pointer to the variable; if it's a bool, get a bitmask too */
		D1(printf("%s: %-20s: call jobs->get_symbol()\n", prompt, __FUNCTION__));

		if (get_symbol (jv_server, &var, reg_cmd->symbol) < 0) {
			res = -EINVAL;
			break;
		}
		/* for byte arrays get the arraysize, too */
		if (var.inttype == jv_bytearray)
			var.cmp.bp = malloc ((size_t)var.arraylen);
		/* for byte arrays get the arraysize, too */
		if (var.inttype == jv_shortarray)
			var.cmp.sp = (uint16_t*)malloc ((size_t)var.arraylen * sizeof(uint16_t));
		/* for byte arrays get the arraysize, too */
		if (var.inttype == jv_intarray)
			var.cmp.ip = (uint32_t*)malloc ((size_t)var.arraylen * sizeof(uint32_t));
		/* for byte arrays get the arraysize, too */
		if (var.inttype == jv_floatarray)
			var.cmp.fp = (float*)malloc ((size_t)var.arraylen * sizeof(float));
		/* for byte arrays get the arraysize, too */
		if (var.inttype == jv_doublearray)
			var.cmp.dp = (double*)malloc ((size_t)var.arraylen * sizeof(double));
	

		/* add var to end of jv_varlist */
		if (jv_server->jv_varlist == NULL)
			jv_server->jv_varlist = list_init();

		jv_server->jv_varlist = 
			list_add_tail(jv_server->jv_varlist, (void*)&var, sizeof(struct jvisu_var));

		jv_server->jv_registeredvars++;
		break;
	}
	
	/* FIXME: Handle string variable */
	case cmd_unregister_var: {

		struct jvisu_regvar_cmd *reg = (struct jvisu_regvar_cmd *) sd->buf;
		int finished = 0; 
		
		D1(
			printf("%s: %-20s: cmd_unregister_var(%s), idx = %d\n", 
			prompt, __FUNCTION__, reg->symbol, je32_to_cpu (reg->idx))
		);
			
		if (is_empty(jv_server->jv_varlist))
			return -EINVAL; /* FIXME: correct? */
		
		/* remove from list, free memory */
		for (goto_head(&jv_server->jv_varlist); !finished; goto_next(&jv_server->jv_varlist)) {

			struct jvisu_var *var = 
				(struct jvisu_var*)list_get_element_data(jv_server->jv_varlist);
		
			if (is_last(jv_server->jv_varlist))
				finished=1;

			if (var->idx != je32_to_cpu (reg->idx) || var->sd != sd) {
				continue;
			}
			
			jv_server->jv_registeredvars--;
			
			if (var->inttype == jv_bytearray)
				if (var->cmp.bp)
					free(var->cmp.bp);
			if (var->inttype == jv_shortarray)
				if (var->cmp.sp)
					free(var->cmp.sp);
			if (var->inttype == jv_intarray)
				if (var->cmp.ip)
					free(var->cmp.ip);
			if (var->inttype == jv_floatarray)
				if (var->cmp.fp)
					free(var->cmp.fp);
			if (var->inttype == jv_doublearray)
				if (var->cmp.dp)
					free(var->cmp.dp);


			jv_server->jv_varlist = list_del_element(jv_server->jv_varlist);
			break;	/* break here, as it is unique (hopefully...) */
		}
		
		break;
	}
				 
	case cmd_conf_data: {
				    
		struct jvisu_sndgen *dat = (struct jvisu_sndgen *) sd->buf;

		D1(
			printf ("%s: %-20s: cmd_conf_data\n", prompt, __FUNCTION__)
		);
		
		dat->data[(sd->cmdlen - sizeof(struct jvisu_cmdhdr))] = '\0';
#if 0		
		if (jops->save_configs)
			jops->save_configs (dat->hdr.type, dat->data);
#endif
		break;
	}
	
	/* write a variable value */
	case cmd_send_var: {
				   
		struct jvisu_sndvar *dat = (struct jvisu_sndvar *) sd->buf;
		
		D1(
			printf("%s: %-20s: cmd_send_var\n", prompt, __FUNCTION__)
		);

		write_variable(jv_server, dat, sd);
		break;
	}
			   
	default:
		break;
	}

	sd->status = 0;
	sd->bufpoi = 0;	
	sd->cmdlen = 0;

	return res;	
}

/** 
 *	Process requests
 *	
 *	@param sd	jvisu socket description structure
 *	@return		0 on success, else error number
 */
static int process_request (struct jv_server *jv_server, struct jvisu_sock *sd)
{
	int	res; 
	size_t	len = sizeof (int);

	/* FIXME: is this correct ? */
	if (sd->status == STAT_WRITE)
		return 0;

	/* check if socket is alive */
	res = read (sd->sd, &sd->buf[sd->bufpoi], 1);
	if (res <= 0) {
		shutdown_sock(jv_server, sd);
		return 1;
	}
	sd->bufpoi++;
	
	len = sizeof (struct jvisu_cmdhdr);
	
	/* loop until buffer is empty or returns EOF */
	while (1) {	
		if (sd->status == STAT_READ) {
			if (!sd->cmdlen)
				len -= sd->bufpoi;
			else
				len = sd->cmdlen - sd->bufpoi;	
		}
		/* set status to READ */
		sd->status = STAT_READ;	

		/* check for more */
		res = read (sd->sd, &sd->buf[sd->bufpoi], len);
		
		/* handle EOF */
		if (res == EOF)
			return 0;

		/* socket error? */
		if (res < 0) {
			shutdown_sock(jv_server, sd);
			return 1;
		}

		/* adjust buffer pointer */
		sd->bufpoi += res;
		
		/* have we read the commandlength already ? */
		if (!sd->cmdlen) {
			if (sd->bufpoi < sizeof (struct jvisu_cmdhdr))
				continue;
			sd->cmdlen = (unsigned int) sd->buf[2];
			sd->cmdlen |= ((unsigned int) sd->buf[3]) << 8;
			D1 (printf ("%s: %-20s: command: len=%d bytes\n", prompt, __FUNCTION__, sd->cmdlen));
		}
		
		/* have we got all bytes ? */
		if (sd->bufpoi == sd->cmdlen) {
			D1 (
				printf ("%s: %-20s: command: analysing (%d bytes read)\n", 
				prompt, __FUNCTION__, (unsigned int)(sd->cmdlen + sizeof (struct jvisu_cmdhdr)))
			);
			return analyse_command(jv_server, sd);
		}	
	}
	return 0; /* keep gcc happy */
}

/** 
 * 	Update fdset
 */
static void check_fdset (struct jv_server *jv_server)
{
	int	i;

	/* check server socket first */
	if (FD_ISSET(jv_server->jv_active_sock[0].sd, &jv_server->jv_read_fdset))
		do_accept(jv_server, jv_server->jv_active_sock[0].sd);
		
	/* server is always set */
	FD_SET(jv_server->jv_active_sock[0].sd, &jv_server->jv_read_fdset);

	/* loop through active connections
	 * 0 is always the server socket
	 */
	for (i = 1; i <= MAX_CONNECTIONS; i++) {

		if (jv_server->jv_active_sock[i].sd == -1) 
			continue;
		
		if (jv_server->jv_active_sock[i].status == STAT_WRITE) {
			if (FD_ISSET(jv_server->jv_active_sock[i].sd, &jv_server->jv_write_fdset)) {
				jv_server->jv_active_sock[i].status &= ~STAT_WRITE;
			}
			FD_SET (jv_server->jv_active_sock[i].sd, &jv_server->jv_write_fdset);
		}
			
		if (FD_ISSET(jv_server->jv_active_sock[i].sd, &jv_server->jv_read_fdset)) {
			if (process_request (jv_server, &jv_server->jv_active_sock[i]))
				continue;
		}
		FD_SET (jv_server->jv_active_sock[i].sd, &jv_server->jv_read_fdset);
	}
}


/**
 *	start/stop JVisu
 *
 *	@param stop	set to 1 if you want to stop JVisu
 *	@return		0 on success
 */
int pv_stop_jvserver (struct jv_server *jv_server, int stop)
{
	jv_server->jvisu_stop = stop;
	return 0;
}


/*
 *  Give JVisu pointers to a process variable
 */
static int get_symbol (struct jv_server *jv_server, struct jvisu_var *var, char *symbol)
{
	struct pv *pv;

	D1(printf("%s: %-20s: JVisu access to symbol <%s>\n", prompt, __FUNCTION__, symbol));

	if (!var || !symbol) {
		printf ("%s: %-20s: error: JVisu access failed\n", prompt, __FUNCTION__);
		return -EINVAL;
	}

	/*
	 * Roadmap: 
	 *
	 * * return -EINVAL if access is invalid
	 *
	 * * val->permissions = JVISU_PERM_RW | JVISU_PERM_RO | JVISU_PERM_WO
	 * * var->data        = (void *)(pointer_to_cell)
	 * * var->inttype     = jv_int, jv_short, ... && return 0
	 *                      return -EINVAL for illegal var type
	 * 
	 */
	
	var->data = NULL;

	pv = pv_register_lookup(jv_server->pv_engine, symbol);
	if (!pv)
		return -EINVAL;

	/* gotcha! */
	D1( printf("%s: %-20s: gotcha! type=%i\n", prompt, __FUNCTION__, pv->type) );
			
	switch (pv->type) {
	case PV_TYPE_INT8:
	case PV_TYPE_UINT8:
		var->inttype = jv_byte;
		break; 
	case PV_TYPE_INT16:
	case PV_TYPE_UINT16:
		var->inttype = jv_short;
		break;
	case PV_TYPE_INT32:
	case PV_TYPE_UINT32:
		var->inttype = jv_int; 
		break;
	case PV_TYPE_STRING:
		var->inttype = jv_string;
		break;
	case PV_TYPE_FLOAT:
		var->inttype = jv_float;
		break;
	case PV_TYPE_DOUBLE:
		var->inttype = jv_double;
		break;
	case PV_TYPE_INT8_ARRAY:
	case PV_TYPE_UINT8_ARRAY:
		var->inttype = jv_bytearray;
		break;
	case PV_TYPE_INT16_ARRAY:
	case PV_TYPE_UINT16_ARRAY:
		var->inttype = jv_shortarray;
		break;
	case PV_TYPE_INT32_ARRAY:
	case PV_TYPE_UINT32_ARRAY:
		var->inttype = jv_intarray;
		break;
	case PV_TYPE_FLOAT_ARRAY:
		var->inttype = jv_floatarray;
		break;
	case PV_TYPE_DOUBLE_ARRAY:
		var->inttype = jv_doublearray;
		break;
	default:
		return -EINVAL;
	}
	
	D1( printf("%s: %-20s: converted to type=%i\n", prompt, __FUNCTION__, var->inttype) );
		
	var->arraylen = pv->arraylen;
	var->data = pv_to_abs_addr(jv_server->pv_engine, pv)->ptr;
	var->permissions = JVISU_PERM_RW; 
	
	return 0;
}


void *jv_server_socket_task (void *arg)
{
	struct jv_server	*jv_server = (struct jv_server *)arg;
	struct timeval		timeo;
	int			res, fd, current = 0;


	/* initialize timekeeping */
	clock_gettime(CLOCK_MONOTONIC, &jv_server->jv_tact);
	
	while (!jv_server->jvisu_stop) {

		/* check if there are active connections */
		fd = 0;
		for (current = 0; current <= MAX_CONNECTIONS; current++) {
			if (jv_server->jv_active_sock[current].sd != -1)
				fd = MAX(jv_server->jv_active_sock[current].sd, fd);
		}
		
		/* 
		 * if no connections -> come back once a second to check 
		 * pvengine->jvisu_stop; else use the given scan interval 
		 * for variable changes
		 */
		timeo.tv_sec = jv_server->jv_connections ? 0 : 1;
		timeo.tv_usec = jv_server->jv_connections ? (JVISU_INTERVAL * 1000) : 0;
		res = select(fd + 1, &jv_server->jv_read_fdset, &jv_server->jv_write_fdset, NULL, &timeo);

		if (res < 0) {
			if (errno == EINTR) continue;
			fprintf (stderr, "%s: %s: error (errno=%i) on select: %d\n", prompt, __FUNCTION__, errno, res);
			continue;			/* FIXME: can we really continue ? */
		}

		/* check fdsets */
		check_fdset(jv_server);				

		/* check variables */
		check_variables(jv_server);
	}

	D1(
		printf ("%s: %-20s: server stopped, shut down sockets\n", 
			prompt, __FUNCTION__)
	);	

	for (current = 0; current <= MAX_CONNECTIONS; current++) {
		if (jv_server->jv_active_sock[current].sd != -1) 
			shutdown_sock (jv_server, &jv_server->jv_active_sock[current]);
	}		

	return NULL;
}

/* end of file jv_server.c */
